#ifndef GPUPPUR_CPU_MESH_ACCELERATORS_BASE_MESH_HPP
#define GPUPPUR_CPU_MESH_ACCELERATORS_BASE_MESH_HPP

/**
 *	@file	
 *	@brief	Base Mesh class of CPU Raytracer class
 *	@author	Tomohiro Matsumoto
 */

#include <gpuppur/utility/begin_suppress_warnings_from_others_code.hpp>
#include <vector>
#include <boost/mpl/int.hpp>
#include <boost/ref.hpp>
#include <boost/iterator/permutation_iterator.hpp>
#include <gpuppur/utility/end_suppress_warnings.hpp>

#include <gpuppur/3dmath/axis_aligned_bounding_box.hpp>

#define PRINT_VAL(n) std::cout << #n << "=" << n << std::endl

namespace gpuppur
{
namespace cpu_raytracers
{
namespace detail
{

/**
 *	@brief	Mesh of CPU Raytracer class
 *
 */
template<bool IsSharedVertices=false, bool IsSharedTriangles=false>
struct base_mesh
{
template<class Parent>
class c
{
public:

	typedef	typename base_mesh<IsSharedVertices>::template c<Parent>	this_type;

private:

	Parent& get_this()
	{
		return *static_cast<Parent*>(this);
	}

	const Parent& get_this() const
	{
		return *static_cast<const Parent*>(this);
	}

public:

	const std::vector<vector3>& get_vertices() const
	{
		return this->vertices;
	}

	const std::vector<unsigned short>& get_triangles(boost::mpl::int_<2>) const
	{
		return this->triangles_16bits;
	}

	const std::vector<unsigned int>& get_triangles(boost::mpl::int_<4>) const
	{
		return this->triangles_32bits;
	}

	const vector3& get_vertex_from_index_of_index
	(
		boost::mpl::int_<2>,
		std::size_t i
	) const
	{
		return this->get_vertices()[this->triangles_16bits[i]];
	}

	const vector3& get_vertex_from_index_of_index
	(
		boost::mpl::int_<4>,
		std::size_t i
	) const
	{
		return this->get_vertices()[this->triangles_32bits[i]];
	}

	const vector3& get_vertex_from_index_of_index
	(
		std::size_t i
	) const
	{
			return this->bytes_per_index==4
		?
			this->get_vertex_from_index_of_index(boost::mpl::int_<4>(), i)
		:
			this->get_vertex_from_index_of_index(boost::mpl::int_<2>(), i);
	}

	const unsigned short get_index
	(
		boost::mpl::int_<2>,
		std::size_t i
	) const
	{
		return this->triangles_16bits[i];
	}

	const unsigned int get_index
	(
		boost::mpl::int_<4>,
		std::size_t i
	) const
	{
		return this->triangles_32bits[i];
	}

	const unsigned int get_index
	(
		std::size_t i
	) const
	{
			return this->bytes_per_index==4
		?
			this->get_index(boost::mpl::int_<4>(), i)
		:
			static_cast<unsigned int>(this->get_index(boost::mpl::int_<2>(), i));
	}

	const std::size_t get_num_indices(boost::mpl::int_<2>) const
	{
		return this->triangles_16bits.size();
	}

	const std::size_t get_num_indices(boost::mpl::int_<4>) const
	{
		return this->triangles_32bits.size();
	}

	const std::size_t get_num_indices() const
	{
			return this->bytes_per_index==4
		?
			this->get_num_indices(boost::mpl::int_<4>())
		:
			this->get_num_indices(boost::mpl::int_<2>());
	}

	const gpuppur::aabb<vector3>&	get_aabb() const
	{
		return this->mesh_aabb;
	}

	const int get_bytes_per_index() const
	{
		return this->bytes_per_index;
	}

public:
/*
	c()
	{
	}
*/
	c
	(
		const std::vector<vector3>&			vertices,
		const std::vector<unsigned short>&	triangles
	):
		bytes_per_index(2),
		vertices(vertices),
		triangles_16bits(triangles),
		triangles_32bits(this_type::null_value(boost::mpl::int_<4>())),
		mesh_aabb
		(
			get_axis_aligned_bounding_box
			(
			//	typename index_buf_itr<unsigned short>::type(vertices.begin(), triangles_16bits.begin()),
			//	typename index_buf_itr<unsigned short>::type(vertices.end(), triangles_16bits.end())
			//	vertices.begin(),
			//	vertices.end()
			boost::make_permutation_iterator(vertices.begin(), triangles.begin()),
			boost::make_permutation_iterator(vertices.end(), triangles.end())
			)
		)
	{
	}

	c
	(
		const std::vector<vector3>&			vertices,
		const std::vector<unsigned int>&	triangles
	):
		bytes_per_index(4),
		vertices(vertices),
		triangles_16bits(this_type::null_value(boost::mpl::int_<2>())),
		triangles_32bits(triangles),
		mesh_aabb
		(
			get_axis_aligned_bounding_box
			(
			//	typename index_buf_itr<unsigned int>::type(vertices.begin(), triangles_32bits.begin()),
			//	typename index_buf_itr<unsigned int>::type(vertices.end(), triangles_32bits.end())
			//	vertices.begin(),
			//	vertices.end()
			boost::make_permutation_iterator(vertices.begin(), triangles.begin()),
			boost::make_permutation_iterator(vertices.end(), triangles.end())
			)
		)
	{
	}

	void operator = (const this_type&)
	{
		assert(false && "Die fucking buggy complier!");
	}

	bool cast_ray_to_aabb
	(
		const ray&	casted_ray,
		const float	max_t
	) const
	{
	#if 1
		boost::tuple<float, float, int, int>
	//	auto	//If you are using compiler	support C++0x
		ret	= gpuppur::cast_ray_to_aabb
		(
			this->get_aabb(),
			casted_ray,
			0.0f,
			max_t
		);

		if(boost::get<2>(ret) == -1 && boost::get<3>(ret) == -1)
		{
			return false;
		}
	#endif

		return true;
	}

	bool cast_ray
	(
		const ray&	casted_ray,
		float&		max_t,				//in:max value of t, out:value of t in hit point
		vector3&	pos,
		vector3&	normal
	) const
	{
		if(this->bytes_per_index==4)
		{
			return get_this().cast_ray_impl
			(
				casted_ray, max_t, pos, normal, boost::mpl::int_<4>()
			);
		}else if(this->bytes_per_index==2)
		{
			return get_this().cast_ray_impl
			(
				casted_ray, max_t, pos, normal, boost::mpl::int_<2>()
			);
		}

		assert(false && "You are using invaild bit index");
		return false;
	}

	bool cast_ray
	(
		const ray&	casted_ray,
		float&		max_t
	) const
	{
		//Optimize someday...>^)))<<

		vector3 pos, normal;

		return this->cast_ray(casted_ray, max_t, pos, normal);
	}

	//Specify shared vertices.
	bool cast_ray
	(
		const ray&	casted_ray,
		float&		max_t,				//in:max value of t, out:value of t in hit point
		vector3&	pos,
		vector3&	normal,
		const std::vector<vector3>& shared_vertices
	) const
	{
		new((void*)(&this->vertices)) boost::reference_wrapper<std::vector<vector3> const>(shared_vertices);

		return this->cast_ray
		(
			casted_ray,
			max_t,
			pos,
			normal
		);
	}

	void analyze_data()
	{
	}

	void draw_analyze_data()
	{
		const gpuppur::VectorNd<vector3, 2>
		box
		(
			this->get_aabb().get_min(),
			this->get_aabb().get_max()
		);

		for(int i=0; i<2; ++i)
		{
			glBegin(GL_LINE_LOOP);
				glVertex3f(box[i][0], box[0][1], box[0][2]);
				glVertex3f(box[i][0], box[0][1], box[1][2]);
				glVertex3f(box[i][0], box[1][1], box[1][2]);
				glVertex3f(box[i][0], box[1][1], box[0][2]);
			glEnd();
		}

		glBegin(GL_LINES);
		for(int i=0; i<2; ++i)
		for(int j=0; j<2; ++j)
		for(int k=0; k<2; ++k)
		{
			glVertex3f(box[k][0], box[j][1], box[i][2]);
		}
		glEnd();
	}

	static std::vector<unsigned short> null_value(boost::mpl::int_<2>)
	{
		return std::vector<unsigned short>();
	}

	static std::vector<unsigned int> null_value(boost::mpl::int_<4>)
	{
		return std::vector<unsigned int>();
	}

private:
/*
	template<typename IndexType>
	struct index_buf_itr
	{
		typedef	boost::permutation_iterator
		<
			typename std::vector<vector3>::iterator,
			typename std::vector<IndexType>::iterator
		>
		type;
	};
*/
	int								bytes_per_index;

	typename boost::mpl::if_c
	<
		IsSharedVertices,
	//	const std::vector<vector3>&,
		boost::reference_wrapper<std::vector<vector3> const>,
		std::vector<vector3>
	>::type							vertices;

	typename boost::mpl::if_c
	<
		IsSharedTriangles,
		const std::vector<unsigned short>&,
		std::vector<unsigned short>
	>::type							triangles_16bits;
	typename boost::mpl::if_c
	<
		IsSharedTriangles,
		const std::vector<unsigned int>&,
		std::vector<unsigned int>
	>::type							triangles_32bits;

	gpuppur::aabb<vector3>			mesh_aabb;

};
};	// end of template class base_mesh


}	// end of namespace detail
}	// end of namespace cpu_raytracers
}	// end of namespace gpuppur

#endif

