#ifndef GPUPPUR_CPU_MESH_ACCELERATORS_OCTREE_HPP
#define GPUPPUR_CPU_MESH_ACCELERATORS_OCTREE_HPP

/**
 *	@file	
 *	@brief	Mesh accelerated by Octree class of CPU Raytracer class
 *	@author	Tomohiro Matsumoto
 */

#include <gpuppur/utility/begin_suppress_warnings_from_others_code.hpp>
#include <algorithm>
#include <vector>
#include <boost/mpl/int.hpp>
#include <gpuppur/utility/end_suppress_warnings.hpp>

#include <gpuppur/3dmath/axis_aligned_bounding_box.hpp>

#include "base_mesh.hpp"

namespace gpuppur
{
namespace cpu_raytracers
{
namespace detail
{

template<template<class> class Base = base_mesh<>::template c>
class octree_mesh : public Base<octree_mesh<Base> >
{
public:

	typedef Base<octree_mesh<Base> >	base;

private:

	template<int BytesPerIndex>
	void divide_mesh
	(
		const std::vector<unsigned int>& triangles,
		const std::size_t octree_index,
		const std::size_t level,
		const std::size_t num_level
	)
	{
		if(level == num_level)
		{
			this->leafs.push_back(triangles);
			return;
		}

		const std::size_t local_child_id = octree_index*8;
		const std::size_t child_id
		=
		this->each_level_begins[level]+local_child_id;

		get_uniform_divided_aabbs
		(
			this->aabb_octree[this->each_level_begins[level-1]+octree_index],
			vector3(2.0f, 2.0f, 2.0f),
			this->aabb_octree.begin()+child_id
		);

		const std::size_t num_triangles = triangles.size();
		for(std::size_t i=0; i<8; ++i)
		{
			std::vector<unsigned int> inner_meshes;
			for(std::size_t j=0; j<num_triangles; ++j)
			{
				for(int k=0; k<3; ++k)
				{
					if
					(
						is_in
						(
							this->aabb_octree[child_id+i],
							base::get_vertex_from_index_of_index
							(
								boost::mpl::int_<BytesPerIndex>(),
								triangles[j]+k
							)
						)
					)
					{
						inner_meshes.push_back(triangles[j]);
						break;
					}
				}
			}

			this->has_geometry_octree[child_id+i] = inner_meshes.size() != 0;
		//	if(!this->has_geometry_octree[child_id+i])
		//		continue;

			this->divide_mesh<BytesPerIndex>
			(
				inner_meshes, local_child_id+i, level+1, num_level
			);
		}
	}

	template<int BytesPerIndex>
	void divide_root_mesh(std::size_t num_level)
	{
		const std::size_t num_indices = base::get_num_indices();

		get_uniform_divided_aabbs
		(
			base::get_aabb(),
			vector3(2.0f, 2.0f, 2.0f),
			this->aabb_octree.begin()
		);

		for(std::size_t i=0; i<8; ++i)
		{
			std::vector<unsigned int> inner_meshes;
			for(std::size_t j=0; j<num_indices; j+=3)
			{
				//buggy code...but it seems a little fine...?
				if
				(
					is_in
					(
						this->aabb_octree[i],
						base::get_vertex_from_index_of_index
						(
							boost::mpl::int_<BytesPerIndex>(),
							j+0
						)
					)
					||
					is_in
					(
						this->aabb_octree[i],
						base::get_vertex_from_index_of_index
						(
							boost::mpl::int_<BytesPerIndex>(),
							j+1
						)
					)
					||
					is_in
					(
						this->aabb_octree[i],
						base::get_vertex_from_index_of_index
						(
							boost::mpl::int_<BytesPerIndex>(),
							j+2
						)
					)
				)
				{
					inner_meshes.push_back(j);
				}
			}

			this->has_geometry_octree[i] = inner_meshes.size() != 0;
		//	if(!this->has_geometry_octree[i])
		//		continue;

			this->divide_mesh<BytesPerIndex>(inner_meshes, i, 1, num_level);
		}
	}

	void gen_octree()
	{
		const std::size_t num_level = 4;
		assert(num_level < 32/3);
		const std::size_t num_last_level_aabb = 1<<(num_level*3); //8^num_level==2^(num_level*3);
		this->aabb_octree.resize(((num_last_level_aabb<<1)-1)&0x49249248);
		this->has_geometry_octree.resize(this->aabb_octree.size());
		this->each_level_begins.resize(num_level);

		std::size_t begin = 0;
		for(std::size_t i=0; i<num_level; ++i)
		{
			this->each_level_begins[i] = begin;
			begin += 1<<((i+1)*3);
		}

		base::get_bytes_per_index()==4
		?
		this->divide_root_mesh<4>(num_level) : this->divide_root_mesh<2>(num_level);
	}


	template<int BytesPerIndex>
	bool cast_ray_to_leaf
	(
		const ray&			casted_ray,
		float&				max_t,
		vector3&			pos,
		vector3&			normal,
		const std::size_t	leaf_index,
		boost::mpl::int_<BytesPerIndex>
	) const
	{
		float min_t = std::numeric_limits<float>::max();

		const std::size_t num_triangles = this->leafs[leaf_index].size();
		for(std::size_t i=0; i<num_triangles; ++i)
		{
			vector3 tmp_pos, tmp_normal;
			float t =
			gpuppur::triangle<vector3>::get_cross
			(
				casted_ray.getPos(),
				casted_ray.getDir(),
				base::get_vertex_from_index_of_index
				(
					boost::mpl::int_<BytesPerIndex>(), this->leafs[leaf_index][i]+0
				),
				base::get_vertex_from_index_of_index
				(
					boost::mpl::int_<BytesPerIndex>(), this->leafs[leaf_index][i]+1
				),
				base::get_vertex_from_index_of_index
				(
					boost::mpl::int_<BytesPerIndex>(), this->leafs[leaf_index][i]+2
				),
				tmp_pos, tmp_normal
			);

			if(t < 0.0f || t > max_t || t >= min_t)
			{
				continue;
			}

			min_t = t;
			pos = tmp_pos;
			normal = tmp_normal;
		}

		if(min_t > max_t)
		{
			return false;
		}

		max_t = min_t;

		return true;
	}

	//I want to define struct local_box in member function cast_ray_to_octree.
	//But it cause link error in visual c++ 2005.
	struct local_box
	{
		std::size_t index;
		float		t_min;

		inline local_box(){}

		inline local_box(std::size_t index, float t_min):
			index(index), t_min(t_min)
		{
		}

		inline bool operator < (const local_box& rhs) const
		{
			return this->t_min < rhs.t_min;
		}
	};

	template<int BytesPerIndex>
	bool cast_ray_to_octree
	(
		const ray&			casted_ray,
		float&				max_t,
		vector3&			pos,
		vector3&			normal,
		const std::size_t	level_local_octree_index,
		const std::size_t	level,
		boost::mpl::int_<BytesPerIndex>
	) const
	{
		if(level == this->each_level_begins.size())
		{
			return
			cast_ray_to_leaf
			(
				casted_ray, max_t, pos, normal,
				level_local_octree_index,
				boost::mpl::int_<BytesPerIndex>()
			);
		}

		const std::size_t local_child_id = level_local_octree_index*8;
		const std::size_t child_id
		=
		this->each_level_begins[level]+local_child_id;

		std::size_t	num_crossed = 0;
		local_box local_boxes[8];
		for(std::size_t i=0; i<8; ++i)
		{
			const std::size_t crnt_id = child_id+i;
			if(!this->has_geometry_octree[crnt_id])
			{
				continue;
			}
		
			boost::tuple<float, float, int, int>
		//	auto	//If you are using compiler	support C++0x
			ret	= gpuppur::cast_ray_to_aabb
			(
				this->aabb_octree[crnt_id],
				casted_ray,
				0.0f, max_t
			);

			const std::size_t crnt_local_id = local_child_id+i;

			if(boost::get<2>(ret) == -1 && boost::get<3>(ret) == -1)
			{
				if
				(
					is_in(this->aabb_octree[crnt_id], casted_ray.getPos(0.0f))
					&&
					is_in(this->aabb_octree[crnt_id], casted_ray.getPos(max_t))
				)
				{
					return 
						cast_ray_to_octree
						(
							casted_ray, max_t, pos, normal,
							crnt_local_id , level+1,
							boost::mpl::int_<BytesPerIndex>()
						);
				}

				continue;
			}else if(boost::get<3>(ret) != -1)
			{
				local_boxes[num_crossed++] = local_box(crnt_local_id, 0.0f);

				continue;
			}

			local_boxes[num_crossed++] = local_box(crnt_local_id, boost::get<0>(ret));
		}

		std::sort(local_boxes+0, local_boxes+num_crossed);

		for(std::size_t i=0; i<num_crossed; ++i)
		{
			if
			(
				this->cast_ray_to_octree
				(
					casted_ray, max_t, pos, normal,
					local_boxes[i].index, level+1,
					boost::mpl::int_<BytesPerIndex>()
				)
			)
			{
				return true;
			}
		}

		return false;
	}

	template<int BytesPerIndex>
	bool cast_ray_impl
	(
		const ray&			casted_ray,
		float&				max_t,
		vector3&			pos,
		vector3&			normal,
		boost::mpl::int_<BytesPerIndex>
	) const
	{
		if(!base::cast_ray_to_aabb(casted_ray, max_t))
		{
			return false;
		}

		return this->cast_ray_to_octree
		(
			casted_ray,
			max_t,
			pos,
			normal,
			0,
			0,
			boost::mpl::int_<BytesPerIndex>()
		);
	}

public:
/*
	octree_mesh()
	{
	}
*/
	template<typename IndexType>
	octree_mesh
	(
		const std::vector<vector3>&			vertices,
		const std::vector<IndexType>&		triangles
	):
		base(vertices, triangles)
	{
		this->gen_octree();

	//	this->analyze_data();
	}

	struct analyze_inner
	{
		static bool pred(const std::vector<unsigned int>& v)
		{
			return v.size() == 0;
		}

		static bool comp(const std::vector<unsigned int>& lhs, const std::vector<unsigned int>& rhs)
		{
			return lhs.size() < rhs.size();
		}
	};

	void analyze_data()
	{
	/*	std::size_t num_node = 8;
		for(std::size_t i=0; i<this->each_level_begins.size(); ++i)
		{
			for(std::size_t j=0; j<num_node; ++j)
			{
				std::cout << this->has_geometry_octree[this->each_level_begins[i]+j] << ',';
			}

			std::cout << std::endl;
			num_node *= 8;
		}
	*/
		std::size_t num_void_leafs
		=
		std::count_if(this->leafs.begin(), this->leafs.end(), &analyze_inner::pred);
		std::size_t max_num_geom
		=
		std::max_element(this->leafs.begin(), this->leafs.end(), &analyze_inner::comp)->size();

		PRINT_VAL(this->leafs.size());
		PRINT_VAL(num_void_leafs);
		PRINT_VAL(max_num_geom);
	}

private:
/*
	class node
	{
		typedef aabb<vector3>
		aabb<vector3>	bound;
		bool			has_geometry;

	public:

		node(const aabb<vector3>& bound, const bool has_geometry):
			bound(bound), has_geometry(has_geometry)
		{
		}

		const aabb<vector3> get_aabb() const
		{
			return this->bound;
		}

		bool get_has_geometry() const
		{
			return this->has_geometry;
		}
	};
*/

	std::vector<aabb<vector3> >					aabb_octree;
	std::vector<bool>							has_geometry_octree;
	std::vector<std::size_t>					each_level_begins;
	std::vector<std::vector<unsigned int> >		leafs;

	friend class base;
};

}	// end of namespace detail
}	// end of namespace cpu_raytracers
}	// end of namespace gpuppur

#endif

