#ifndef UTILITY_PROFILER
#define UTILITY_PROFILER

/**
 *	@file	
 *	@brief	Class for profiling.
 *	@author	Tomohiro Matsumoto
 */

#include <map>
#include <boost/foreach.hpp>

namespace gpuppur
{
#ifdef WIN32
typedef __int64 clock_type;

inline clock_type get_clock()
{
	__asm
	{
		rdtsc
	}
}
#endif

class database
{
	class record
	{
	public:

		record()
		{
			assert(false && "This function is needed to compile, but it will never be called");
		}

		record(std::size_t index, clock_type total, const char* name):
			index(index), total(total), noise(0), num_called(1), name(name)
		{
		}

		void add(clock_type time, clock_type noise)
		{
			this->total += time;
			this->noise += noise;
			++this->num_called;
		}

		void add_noise(clock_type time)
		{
			this->noise += time;
		}

		bool operator < (const record& rhs) const
		{
			return this->index < rhs.index;
		}

		void print()
		{
			std::cout << this->total << '/' << this->num_called << ':' << this->name << std::endl;
		}

	private:

		std::size_t	index;
		clock_type	total;
		clock_type	noise;
		std::size_t num_called;
		const char*	name;

	};

	typedef std::map<const char*, record>	record_type;

public:

	void add(clock_type time, const char* name)
	{
		clock_type noise_begin = get_clock();
		if(this->records.find(name) == this->records.end())
		{
			this->records.insert
			(
				record_type::value_type
				(
					name, record(this->records.size(), time, name)
				)
			);

			this->records[name].add_noise(get_clock()-noise_begin);

			return;
		}

		this->records[name].add(time, get_clock()-noise_begin);
	}

	void print() const
	{
		std::vector<record> sorted;

		sorted.reserve(this->records.size());
		record_type::const_iterator i = this->records.begin();
		for(; i!= this->records.end(); ++i)
		{
			sorted.push_back(i->second);
		}

		std::sort(sorted.begin(), sorted.end());

		BOOST_FOREACH(record& r, sorted)
		{
			r.print();
		}
	}

	void clear()
	{
		this->records.clear();
	}

private:

	record_type 	records;
};

inline database& get_database()
{
	static database d;

	return d;
}

class profiler
{
public:

	profiler(const char* name):
		name(name)
	{
		this->begin = get_clock();
	}

	~profiler()
	{
		get_database().add(get_clock() - this->begin, name);
	}

private:

	const char*	name;
	clock_type	begin;

};

}	//end of namespace gpuppur

#endif
